class ListStoreSyncer(object):

    list1 = None
    list2 = None

    def __init__(self, list1, list2, map_from_list1, map_from_list2):
        self.map = {
            list1: list2,
            list2: list1,
        }
        self.map_value = {
            list1: map_from_list1,
            list2: map_from_list2,
        }
        self.watches = {
            list1: [
                list1.connect('row-changed', self.on_row_changed),
                list1.connect('row-deleted', self.on_row_deleted),
                list1.connect('row-inserted', self.on_row_inserted),
                list1.connect('rows-reordered', self.on_rows_reordered),
            ],
            list2: [
                list2.connect('row-changed', self.on_row_changed),
                list2.connect('row-deleted', self.on_row_deleted),
                list2.connect('row-inserted', self.on_row_inserted),
                list2.connect('rows-reordered', self.on_rows_reordered),
            ]
        }

    def close(self):
        for lst, watches in self.watches.items():
            [lst.disconnect(w) for w in watches]

    def on_row_changed(self, source, path, iter_, *unused):
        dest = self.map[source]
        map_ = self.map_value[source]
        new_row = map_([x for x in source[iter_]])
        original_row = [x for x in dest[dest.get_iter(path)]]
        if new_row != original_row:
            dest.handler_block(self.watches[dest][0])
            dest[dest.get_iter(path)] = new_row
            dest.handler_unblock(self.watches[dest][0])

    def on_row_deleted(self, source, path, *unused):
        dest = self.map[source]
        dest.handler_block(self.watches[dest][1])
        dest.remove(dest.get_iter(path))
        dest.handler_block(self.watches[dest][1])

    def on_row_inserted(self, source, path, iter_, *unused):
        dest = self.map[source]
        map_ = self.map_value[source]
        new_row = map_([x for x in source[iter_]])
        source.handler_block(self.watches[source][2])
        dest.handler_block(self.watches[dest][2])
        dest.insert(path[0], new_row)
        dest.handler_unblock(self.watches[dest][2])
        source.handler_unblock(self.watches[source][2])

    def on_rows_reordered(self, source, *unused):
        dest = self.map[source]
        dest.handler_block(self.watches[dest][3])
        raise NotImplementedError
        dest.handler_unblock(self.watches[dest][3])



#!/usr/bin/python3

import logging

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk

import simplegtd.models.blobstore
import simplegtd.models.linestore
import simplegtd.models.taskstore
import simplegtd.models.liststoresyncer
from simplegtd.task import markup_for


# FIXME: using some pygtk magic, eliminate the need for this additional sync
# between the taskstore and the todotxt (effectively eliminating this one
# object for good, or paring it down to a stub), by using cellrenderers
# that use markup generated directly from the Task object, because only the
# task object knows how to adequately identify the tokens present in the task
# text.
class TodoTxt(Gtk.ListStore, simplegtd.models.liststoresyncer.ListStoreSyncer):

    filename = None
    blobstore = None
    linestore = None
    taskstore = None

    __gsignals__ = {
        'ready': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, filename):
        Gtk.ListStore.__init__(self, object, str)
        self.filename = filename
        self.logger = logging.getLogger(self.__class__.__name__)
        self.blobstore = simplegtd.models.blobstore.FileBlobStore()
        self.linestore = simplegtd.models.linestore.LineStore()
        self.__watches = [
            self.blobstore.connect("done-reading", self.linestore.unserialize),
        ]
        self.__ready_watches = [
            self.linestore.connect("unserialized", self.__on_first_unserialize),
        ]
        self.taskstore = simplegtd.models.taskstore.TaskStore(self.linestore)
        simplegtd.models.liststoresyncer.ListStoreSyncer.__init__(
            self, self, self.taskstore,
            lambda r: [r[0]], lambda r: [r[0], markup_for(r[0].text)],
        )

    def __on_first_unserialize(self, *unused):
        # Only fires the first time that the linestore is unserialized.
        self.linestore.disconnect(self.__ready_watches[-1])
        self.__ready_watches = self.__ready_watches[:-1]
        self.logger.debug("Successful unserialize of the linestore.  Emitting ready().")
        self.emit('ready')

    def name(self):
        return self.filename

    def open(self):
        return self.blobstore.open(self.filename)

    def close(self):
        simplegtd.models.liststoresyncer.ListStoreSyncer.close(self)
        self.taskstore.close()
        [self.linestore.disconnect(w) for w in self.__ready_watches]
        [self.blobstore.disconnect(w) for w in self.__watches]
        self.blobstore.close()

    def remove_at(self, iter_):
        Gtk.ListStore.remove(self, iter_)
        self.__save()

    def new_at(self, iter_):
        if iter_ is not None:
            path = self.get_path(iter_)
            row = path[0]
        else:
            row = 0
        new_iter = Gtk.ListStore.insert(self, row, [simplegtd.models.taskstore.Task.fromstring(""), ""])
        return new_iter

    def get_at(self, iter_):
        return self.get_value(iter_, 0).text

    def remove_many(self, iters):
        removed = False
        for iter_ in iters:
            removed = True
            Gtk.ListStore.remove(self, iter_)
        if removed:
            self.__save()

    def edit(self, iter_, new_text):
        '''Edit task at iter_.  Does nothing if the line does not change.'''
        if self.get_value(iter_, 0).text == new_text:
            return
        self[iter_][0].set_text(new_text)
        self[iter_][1] = markup_for(new_text)
        self.__save()

    def __save(self):
        '''Saves the todo tasks list.'''
        text = self.linestore.serialize()
        # FIXME errors in puts must be handled somehow.
        # Right now they are not handled.
        self.blobstore.put(text)



#!/usr/bin/python3

__version__ = "0.0.25"


import argparse
import collections
import logging
import os

import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from gi.repository import GObject, GLib, Gdk, Gtk, Gio
import simplegtd.libwhiz.path
import simplegtd.mainwindow
import simplegtd.todotxt




class _SimpleGTDAppState(object):

    todo_txts = None
    __app_state_file = None

    def __init__(self, filename):
        self.__app_state_file = filename
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(filename, GLib.KeyFileFlags.NONE)
            self.todo_txts = f.get_string_list("general", "todo_txts")
        except GLib.Error:
            pass
        return self

    def persist_app_state(self):
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(self.__app_state_file, GLib.KeyFileFlags.NONE)
        except GLib.Error:
            pass
        if self.todo_txts is not None:
            f.set_string_list("general", "todo_txts", self.todo_txts)
        f.save_to_file(self.__app_state_file)


class SimpleGTD(Gtk.Application, _SimpleGTDAppState):
    def __init__(self):
        Gtk.Application.__init__(self)
        self.config_home = simplegtd.libwhiz.path.config_home("simplegtd")
        self.data_dir = simplegtd.libwhiz.path.data_home("simplegtd")
        self.cache_home = simplegtd.libwhiz.path.cache_home("simplegtd")
        for d in self.config_home, self.data_dir, self.cache_home:
            if not os.path.isdir(d):
                os.makedirs(d)

        settings_file = os.path.join(self.config_home, "app-settings")
        _SimpleGTDAppState.__init__(self, settings_file)
        self.default_todo_txt = os.path.join(self.data_dir, "todo.txt")
        if not self.todo_txts:
            self.todo_txts = [self.default_todo_txt]

        self.connect("activate", self.on_activate)
        self.connect("window-removed", self.on_todo_window_removed)
        self.connect("shutdown", self.on_shutdown)
        self.models_to_windows = {}

    def on_activate(self, *a):
        logging.debug("Activate: %s", a)
        for x in self.todo_txts:
            self.add_new_todo_window(x)

    def delegate_open_file_activated(self, requestor):
        choosefile_dialog = Gtk.FileChooserDialog(
            title="Select an existing TODO.TXT file or create a new one",
            parent=requestor,
            action=Gtk.FileChooserAction.SAVE,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
                "Use default",
                Gtk.ResponseType.NONE,
            ),
        )
        response = choosefile_dialog.run()
        choosefile_dialog.hide()
        if response == Gtk.ResponseType.OK:
            self.add_new_todo_window(choosefile_dialog.get_filename())
        elif response == Gtk.ResponseType.NONE:
            self.add_new_todo_window(self.default_todo_txt)
        choosefile_dialog.destroy()

    def delegate_new_window_activated(self, requestor):
        for data_file, (unused_model, windows) in self.models_to_windows.items():
            if requestor in windows:
                self.add_new_todo_window(data_file)

    def delegate_exit_activated(self, unused_requestor):
        self.quit()

    def add_new_todo_window(self, data_file):
        try:
            if data_file in self.models_to_windows:
                model = self.models_to_windows[data_file][0]
            else:
                model = simplegtd.todotxt.TodoTxt(data_file)
                # FIXME: connect blobstore done-writing and done-reading to handlers
                # that will handle any exception encountered during the operation.
                # FIXME: disable all data entry and editing between reading/done-reading
                # and writing/done-writing.  All fields and actions must be greyed out
                # while that's happening.  If an edit is in progress, still have to decide
                # what we will do in order to handle that situation.  Or decide later how we
                # will handle that circumstance.
                # FIXME: perhaps show window *after* successful first load, or error message
                # if unsuccessful load.  If unsuccessful load, perhaps the window must be
                # destroyed and cleaned up completely.
                model.open()
            window = simplegtd.mainwindow.SimpleGTDMainWindow(
                model,
                os.path.join(
                    self.cache_home,
                    "window-state-" + simplegtd.libwhiz.path.hash_path(data_file),
                ),
            )
            self.add_window(window)
            if data_file not in self.models_to_windows:
                self.models_to_windows[data_file] = (model, [])
            self.models_to_windows[data_file][1].append(window)
            window.connect("open-file-activated", self.delegate_open_file_activated)
            window.connect("new-window-activated", self.delegate_new_window_activated)
            window.connect("exit-activated", self.delegate_exit_activated)
            window.connect("destroy", self.remove_window)
            window.show_all()
        except BaseException:
            Gtk.main_quit()
            raise

    def on_todo_window_removed(self, unused_ref, window):
        for data_file, (model, windows) in list(self.models_to_windows.items()):
            if window in windows:
                windows.remove(window)
            if not windows:
                model.close()
                del self.models_to_windows[data_file]
        if not self.models_to_windows:
            pass
        else:
            self.todo_txts = list(self.models_to_windows.keys())
            self.persist_app_state()

    def on_shutdown(self, *unused):
        logging.debug("We have models-to-windows: %s", self.models_to_windows.keys())
        if self.models_to_windows:
            # First, we disconnect our own on_todo_window_removed handler,
            # since while quitting we do not want that handler to handle
            # whittling down the window list.
            try:
                self.disconnect_by_func(self.on_todo_window_removed)
            except TypeError:
                pass
            # Now we persist everything.
            self.todo_txts = list(self.models_to_windows.keys())
            self.persist_app_state()
            for model, windows in list(self.models_to_windows.values()):
                for window in windows:
                    window.destroy()
            for model, windows in list(self.models_to_windows.values()):
                model.close()
            self.models_to_windows.clear()


parser = argparse.ArgumentParser(description="Simple GTD")
parser.add_argument("-l", "--log-level", default=None, choices=["debug"])
# FIXME: special-case handling of files opened via command-line
# so that files in question can be opened.  This may require the app
# to become a single-instance app, so that the window state can be
# persisted.


def main():
    args = parser.parse_args()
    if args.log_level == "debug":
        logging.basicConfig(level=getattr(logging, args.log_level.upper()))
    app = SimpleGTD()
    app.run()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import collections
import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from simplegtd.task import markup_for


class FilterList(Gtk.ListStore):

    def __init__(self, todotxt):
        Gtk.ListStore.__init__(self, str, str, str, str)
        '''Contains filters to be applied to the task list.

        Columns:
           1. plain text string
           2. sortable representation
           3. filter string
           4. markup to render instead of plain text string
        '''
        self.logger = logging.getLogger(self.__class__.__name__)
        self.tokens_contained_in_row = []
        self.token_visible = collections.defaultdict(bool)
        self.append(["All tasks", "", "", "All tasks"])
        todotxt.connect("row-inserted", self.todotxt_row_inserted_or_changed)
        todotxt.connect("row-changed", self.todotxt_row_inserted_or_changed)
        todotxt.connect("row-deleted", self.todotxt_row_deleted)
        for n, unused_row in enumerate(todotxt):
            self.todotxt_row_inserted_or_changed(todotxt, Gtk.TreePath.new_from_indices([n]), None)

    def todotxt_row_inserted_or_changed(self, todotxt, path, unused_it):
        row = path.get_indices()[0]
        text = todotxt[path][0].text

        # Fill up the list to capacity.
        while len(self.tokens_contained_in_row) <= row:
            self.tokens_contained_in_row.append([])
        # Remember the old tokens before the change.
        old_tokens = self.tokens_contained_in_row[row]
        # Zero out the token store for this row.
        self.tokens_contained_in_row[row] = []

        # The task is empty?  Do nothing then.
        if text is None:
            # Deref old tokens, return.
            self._deref_old_tokens(old_tokens, [])
            return
        else:
            # Deref old tokens ignoring new ones.
            new_tokens = [t.strip() for t in text.split()]
            self._deref_old_tokens(old_tokens, new_tokens)

        # Add tokens that are not visible at the moment.
        for tok in new_tokens:
            if tok.startswith("@") or tok.startswith("+"):
                self.tokens_contained_in_row[row].append(tok)
                if not self.token_visible[tok]:
                    sortable = ("1" if tok.startswith("@") else "2") + tok
                    self.append([tok, sortable, tok, markup_for(tok)])
                    self.token_visible[tok] = True

    def todotxt_row_deleted(self, unused_todotxt, path):
        row = path.get_indices()[0]
        old_tokens = self.tokens_contained_in_row[row]
        # Remove the tokens on the way out.
        self.tokens_contained_in_row.pop(row)
        # Deref old tokens.
        self._deref_old_tokens(old_tokens, [])

    def _deref_old_tokens(self, old_tokens, new_tokens):
        # Compute tokens of remaining tasks, then remove filter
        # elements that do not correspond to tokens of remaining tasks.
        remaining = set(tok for rowtoks in self.tokens_contained_in_row for tok in rowtoks)
        remaining = remaining | set(new_tokens if new_tokens else [])
        for tok in old_tokens:
            if self.token_visible[tok] and tok not in remaining:
                self.token_visible[tok] = False
                for r in self:
                    if r[0] == tok:
                        self.remove(r.iter)
                        break

    def get_sorted(self):
        '''Wrap the FilterList into a sortable list.'''
        sorted_filters = Gtk.TreeModelSort(model=self)
        sorted_filters.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        return sorted_filters

    def get_filter_string(self, iter_):
        '''Return the filter string pointed to by iter_.'''
        return self.get_value(iter_, 2)



import collections
import datetime
import logging
import re

import gi
gi.require_version('GObject', '2.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk

import simplegtd.models.liststoresyncer

RegexCapsule = collections.namedtuple('RegexCapsule', 'regex group')


CREATION_DATE_RE = RegexCapsule(re.compile(r'(^|\s)(\d\d\d\d-[0-1]\d-[0-3]\d)($|\s)'), 2)
THRESHOLD_DATE_RE = RegexCapsule(re.compile(r'(^|\s)t:(\d\d\d\d-[0-1]\d-[0-3]\d)($|\s)'), 2)


def unencap(capsule, text):
    try:
        r = capsule.regex.search(text)
    except TypeError:
        print(repr(text))
        raise
    if r is None:
        return None
    return r.group(capsule.group)


# FIXME: TODO: add unit tests for this stuff
# FIXME: finish task parsing, based on Simpletask task formatting.
class Task(GObject.GObject):

    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    creation_date = None
    threshold_date = None
    text = ""

    def __init__(self):
        GObject.GObject.__init__(self)

    def __str__(self):
        return '<%s>' % self.text

    def __repr__(self):
        return self.__str__()

    @classmethod
    def fromstring(klass, text):
        k = klass()
        k.text = text
        k._update()
        return k

    def set_text(self, new_text):
        if new_text != self.text:
            self.text = new_text
            self._update()
            self.emit("changed")

    def _update(self):
        self._update_creation_date()
        self._update_threshold_date()

    def _unencap(self, capsule):
        return unencap(capsule, self.text)

    def _update_creation_date(self):
        ds = self._unencap(CREATION_DATE_RE)
        if ds:
            try:
                self.creation_date = datetime.datetime.strptime(ds, "%Y-%m-%d").date()
            except ValueError:
                self.creation_date = None
        else:
            self.creation_date = None

    def _update_threshold_date(self):
        ds = self._unencap(THRESHOLD_DATE_RE)
        if ds:
            try:
                self.threshold_date = datetime.datetime.strptime(ds, "%Y-%m-%d").date()
            except ValueError:
                self.threshold_date = None
        else:
            self.threshold_date = None


class TaskStore(Gtk.ListStore, simplegtd.models.liststoresyncer.ListStoreSyncer):

    linestore = None

    def __init__(self, linestore):
        Gtk.ListStore.__init__(self, object)
        simplegtd.models.liststoresyncer.ListStoreSyncer.__init__(
            self, self, linestore,
            lambda r: [r[0].text], lambda r: [Task.fromstring(r[0])],
        )
        self.logger = logging.getLogger(self.__class__.__name__)
        self.linestore = linestore

    def on_task_changed(self, task):
        iter_ = self.get_iter_first()
        while iter_:
            if self[iter_][0] is task:
                break
            iter_ = self.iter_next(iter_)
        if iter_:
            self.emit("row-changed", self.get_path(iter_), iter_)

    def on_row_changed(self, source, path, iter_, *unused):
        if source == self:
            return simplegtd.models.liststoresyncer.ListStoreSyncer.on_row_changed(
                self, source, path, iter_, *unused
            )
        text = source[path][0]
        task = self[path][0]
        if text != task.text:
            self.handler_block(self.watches[self][0])
            task.set_text(text)
            self.handler_unblock(self.watches[self][0])

    def on_row_deleted(self, source, path, *unused):
        if source == self.linestore:
            task = self[path][0]
            task.disconnect_by_func(self.on_task_changed)
        return simplegtd.models.liststoresyncer.ListStoreSyncer.on_row_deleted(
            self, source, path, *unused
        )

    def on_row_inserted(self, source, path, *unused):
        ret = simplegtd.models.liststoresyncer.ListStoreSyncer.on_row_inserted(
            self, source, path, *unused
        )
        self[path][0].connect("changed", self.on_task_changed)
        return ret

    def close(self):
        simplegtd.models.liststoresyncer.ListStoreSyncer.close(self)
        for row in self:
            row[0].disconnect_by_func(self.on_task_changed)



import datetime
import unittest

import gi
gi.require_version('GLib', '2.0')
from gi.repository import GLib

import simplegtd.models.taskstore as ts
import simplegtd.models.linestore as ls
import simplegtd.models.blobstore as bs


class TaskStoreTest(unittest.TestCase):

    def setUp(self):
        self.b = bs.StringBlobStore()
        self.l = ls.LineStore()
        self.b.connect("done-reading", self.l.unserialize)
        self.t = ts.TaskStore(self.l)
        self.c = GLib.MainContext.default()

    def run_loop(self):
        while self.c.pending():
            self.c.iteration(False)

    def tearDown(self):
        del self.c
        self.t.close()
        del self.t
        self.b.disconnect_by_func(self.l.unserialize)
        del self.l
        self.b.close()
        del self.b

    def test_adding_to_blobstore_adds_tasks(self):
        text = "abc 2020-04-03 a task t:2020-05-05"
        self.b.open(text)
        self.run_loop()
        assert len(self.t) == 1
        task = self.t[0][0]
        self.assertEqual(text, task.text)
        self.assertEqual(task.creation_date, datetime.date(2020, 4, 3))
        self.assertEqual(task.threshold_date, datetime.date(2020, 5, 5))

        text2 = "cdef"
        text = text2 + "\n" + text
        # Simulate write to backend, and its side effect.
        self.b.put(text)
        self.b.emit('done-reading', text, None)
        self.run_loop()
        assert len(self.t) == 2, [t[0] for t in self.t]
        task = self.t[0][0]
        self.assertEqual(text2, task.text)
        self.assertEqual(task.creation_date, None)

    def test_removal_from_blobstore_removes_tasks(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        assert len(self.t) == 1
        self.b.put("")
        self.b.emit('done-reading',"", None)
        self.run_loop()
        assert len(self.t) == 0

    def test_change_in_blobstore_changes_task(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        assert len(self.t) == 1
        task = self.t[0][0]
        # Simulate write to backend, and its side effect.
        text = "abc 2020-04-03 a task"
        self.b.put(text)
        self.b.emit('done-reading', text, None)
        self.run_loop()
        self.assertEqual(self.t[0][0].creation_date, datetime.date(2020, 4, 3))
        self.assertEqual(self.t[0][0].threshold_date, None)
        assert self.t[0][0] is task

    def test_adding_task_syncs_to_blobstore(self):
        task = "abc 2020-04-03 a task t:2020-05-05"
        self.b.open("")
        self.run_loop()
        assert len(self.t) == 0
        assert len(self.l) == 0
        self.t.append([ts.Task.fromstring(task)])
        self.run_loop()
        assert len(self.t) == 1
        assert len(self.l) == 1
        self.assertEqual(self.l[0][0], task)

    def test_removing_task_syncs_to_blobstore(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        assert len(self.t) == 1
        assert len(self.l) == 1, (len(self.l), [x[0] for x in self.l])
        self.t.remove(self.t.get_iter_first())
        self.run_loop()
        assert len(self.t) == 0
        assert len(self.l) == 0

    def test_changing_task_object_syncs_to_blobstore(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        changed_task = "abc 2020-04-03 a task"
        self.t[0][0] = ts.Task.fromstring(changed_task)
        self.t[0][0].connect("changed", self.t.on_task_changed)
        self.run_loop()
        self.assertEqual(self.t[0][0].threshold_date, None)
        self.b.put(self.l.serialize())
        self.run_loop()
        self.assertEqual(self.b.memory, changed_task)
        self.t[0][0] = self.t[0][0]
        self.run_loop()

    def test_changing_task_content_syncs_to_blobstore(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        changed_task = "abc 2020-04-03 a task"
        self.t[0][0].set_text(changed_task)
        self.run_loop()
        self.assertEqual(self.t[0][0].threshold_date, None)
        self.b.put(self.l.serialize())
        self.run_loop()
        self.assertEqual(self.b.memory, changed_task)

    def test_deleting_task_syncs_to_blobstore_and_mods_after_deletion_do_not_cause_errors(self):
        self.b.open("abc 2020-04-03 a task t:2020-05-05")
        self.run_loop()
        task = self.t[0][0]
        self.b.put("")
        self.run_loop()
        task.set_text("abc")
        self.run_loop()
        assert self.b.memory == ""



import os

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
try:
    gi.require_version('Handy', '1')
    handy_version = 1
except ValueError:
    handy_version = 0
    gi.require_version('Handy', '0.0')

from gi.repository import GObject, GLib, Gdk, Gtk, Handy

import simplegtd.libwhiz.rememberingwindow
import simplegtd.filterlist
import simplegtd.libwhiz.path
import simplegtd.views


if handy_version == 1:
    FilterHeaderBarClass = Handy.HeaderBar
    TasksHeaderBarClass = Handy.HeaderBar
else:
    FilterHeaderBarClass = Gtk.HeaderBar
    TasksHeaderBarClass = Gtk.HeaderBar


def make_title(todotxt):
    if todotxt.name():
        n = todotxt.name()
        bn = os.path.basename(n)
        dn = simplegtd.libwhiz.path.strip_data_home(os.path.dirname(n))
        if dn:
            dn = simplegtd.libwhiz.path.abbrev_home(dn)
            title = "%s (%s)" % (bn, dn)
        else:
            title = "%s" % bn
        return title
    return 'Simple GTD (in memory)'


class FilterHeaderBar(FilterHeaderBarClass):

    saved_filter_text = ""
    text_filter_mode_active = False

    __gsignals__ = {
        'search-changed': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, (str,)),
        'search-deactivated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'activate': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
    }

    def __init__(self):
        Gtk.HeaderBar.__init__(self)
        self.set_show_close_button(False)
        self.set_title("Filters")
        self.text_filter_entry = Gtk.SearchEntry()
        self.text_filter_entry.set_placeholder_text("Search...")
        self.text_filter_entry.set_no_show_all(True)
        self.pack_start(self.text_filter_entry)
        self.search_button = Gtk.Button.new_from_icon_name("edit-find", Gtk.IconSize.LARGE_TOOLBAR)
        self.search_button.connect("clicked", lambda *_: self.activate_text_filter_mode())
        self.pack_start(self.search_button)
        self.text_filter_entry.connect("search-changed", lambda _: self.emit("search-changed", self.text_filter_entry.get_text()))
        self.text_filter_entry.connect("activate", lambda _: self.emit("activate"))
        self.text_filter_entry.connect("stop-search", lambda *_: self.deactivate_text_filter_mode())
        self.text_filter_entry.connect("focus-out-event", lambda *_: self.filter_unfocused())

    def filter_unfocused(self):
        if self.text_filter_mode_active:
            if self.text_filter_entry.is_visible() and not self.text_filter_entry.get_text().strip():
                self.deactivate_text_filter_mode()

    def activate_text_filter_mode(self):
        if not self.text_filter_mode_active:
            w = Gtk.Label()
            w.set_visible(False)
            self.set_custom_title(w)
            self.search_button.hide()
            self.text_filter_entry.set_text(self.saved_filter_text)
            self.text_filter_entry.show()
        self.text_filter_entry.grab_focus()
        self.text_filter_mode_active = True

    def deactivate_text_filter_mode(self):
        if self.text_filter_mode_active:
            self.text_filter_mode_active = False
            self.set_custom_title(None)
            self.saved_filter_text = self.text_filter_entry.get_text()
            self.text_filter_entry.set_text("")
            self.text_filter_entry.hide()
            self.search_button.show()
            self.emit("search-deactivated")

    def toggle_text_filter_mode(self):
        if self.text_filter_mode_active:
            if self.text_filter_entry.is_focus():
                self.deactivate_text_filter_mode()
            else:
                self.text_filter_entry.grab_focus()
        else:
            self.activate_text_filter_mode()


class SimpleGTDMainWindow(Gtk.ApplicationWindow, simplegtd.libwhiz.rememberingwindow.RememberingWindow):

    __gsignals__ = {
        'open-file-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'new-window-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'close-window-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'text-filter-toggled': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'exit-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'help-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
    }

    filter_header_bar = FilterHeaderBar
    tasks_header_bar = TasksHeaderBarClass

    def __init__(self, todotxt, window_state_file):
        self.selection_filters = []
        self.search_filters = []

        Gtk.ApplicationWindow.__init__(self)
        self.set_default_icon_name("simplegtd")

        self.set_default_size(800, 600)
        simplegtd.libwhiz.rememberingwindow.RememberingWindow.__init__(self, window_state_file)

        accel_group = Gtk.AccelGroup()
        self.add_accel_group(accel_group)

        self.add_accelerator("close-window-activated", accel_group, ord('w'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.connect("close-window-activated", lambda _: self.destroy())

        self.filter_header_bar = self.filter_header_bar()
        self.filter_header_bar.connect("search-changed", lambda _, text: self.filter_text_changed(text))
        self.filter_header_bar.connect("search-deactivated", lambda unused_entry: self.task_view.grab_focus())
        self.filter_header_bar.connect("activate", lambda unused_entry: self.task_view.grab_focus())
        self.add_accelerator("text-filter-toggled", accel_group, ord('f'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.connect('text-filter-toggled', lambda *_: self.filter_header_bar.toggle_text_filter_mode())

        self.tasks_header_bar = self.tasks_header_bar()
        self.tasks_header_bar.set_show_close_button(True)

        [x(make_title(todotxt)) for x in (self.tasks_header_bar.set_title, self.set_title)]
        todotxt.connect("row-inserted", self.recompute_subtitle)
        todotxt.connect("row-deleted", self.recompute_subtitle)
        self.recompute_subtitle(todotxt)

        exit_button = Gtk.Button.new_from_icon_name("application-exit", Gtk.IconSize.LARGE_TOOLBAR)
        exit_button.connect("clicked", lambda _: self.emit("exit-activated"))
        self.add_accelerator("exit-activated", accel_group, ord('q'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.tasks_header_bar.pack_end(exit_button)

        new_view_button = Gtk.Button.new_from_icon_name("window-new", Gtk.IconSize.LARGE_TOOLBAR)
        new_view_button.connect("clicked", lambda _: self.emit("new-window-activated"))
        self.add_accelerator("new-window-activated", accel_group, ord('n'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.tasks_header_bar.pack_end(new_view_button)

        choosefile_button = Gtk.Button.new_from_icon_name("document-open", Gtk.IconSize.LARGE_TOOLBAR)
        choosefile_button.connect("clicked", lambda _: self.emit("open-file-activated"))
        self.add_accelerator("open-file-activated", accel_group, ord('o'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.tasks_header_bar.pack_end(choosefile_button)

        help_button = Gtk.Button.new_from_icon_name("system-help", Gtk.IconSize.LARGE_TOOLBAR)
        help_button.connect("clicked", lambda _: self.show_shortcuts_window())
        self.add_accelerator("help-activated", accel_group, Gdk.KEY_F1, 0, 0)
        self.connect("help-activated", lambda _: self.show_shortcuts_window())
        self.tasks_header_bar.pack_end(help_button)

        self.task_view = simplegtd.views.TaskView()
        todotxt.connect("ready", lambda *unused: self.task_view.focus_first())
        task_view_scroller = Gtk.ScrolledWindow()
        task_view_scroller.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        task_view_scroller.add(self.task_view)

        self.filter_view = simplegtd.views.FilterView()
        GLib.idle_add(self.filter_view.select_first)
        self.filter_view.get_selection().connect("changed", self.filter_selection_changed)
        filter_view_scroller = Gtk.ScrolledWindow()
        filter_view_scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        filter_view_scroller.add(self.filter_view)

        filters = simplegtd.filterlist.FilterList(todotxt)
        self.task_view.set_model(todotxt)
        self.filter_view.set_model(filters)

        # Group the header bars.
        hg = Handy.HeaderGroup()
        hg.add_header_bar(self.filter_header_bar)
        hg.add_header_bar(self.tasks_header_bar)

        # Group the filter header bar and filter scroller list.
        sg = Gtk.SizeGroup()
        sg.set_mode(Gtk.SizeGroupMode.HORIZONTAL)
        sg.add_widget(self.filter_header_bar)
        sg.add_widget(filter_view_scroller)

        # Make the title bar.
        titlebox = Gtk.Box()
        titlebox.pack_start(self.filter_header_bar, False, True, 0)
        self.filter_header_bar.set_show_close_button(True)
        sep = Gtk.VSeparator()
        sep.get_style_context().add_class("sidebar")
        titlebox.add(sep)
        titlebox.pack_end(self.tasks_header_bar, True, True, 0)
        self.tasks_header_bar.set_show_close_button(True)
        ht = Handy.TitleBar()
        ht.add(titlebox)
        self.set_titlebar(ht)

        # Make the pane components and add the pane to the window.
        paned = Handy.Leaflet()
        paned.add(filter_view_scroller)
        paned.add(task_view_scroller)
        paned.set_visible_child(task_view_scroller)
        self.add(paned)

        self.task_view.grab_focus()

    def recompute_subtitle(self, model, *unused_args):
        # FIXME we need to show outstanding tasks rather than all tasks
        # (that requires a richer data model for the tasks stored in the
        # model) and we also require to show which ones are hidden by filters
        # as well as what filter is being used.  Perhaps a "last modified
        # n minutes ago" with a rough time would be useful.
        numtasks = len(model)
        self.tasks_header_bar.set_subtitle("%s tasks" % numtasks)

    def filter_selection_changed(self, tree_selection):
        filter_strings = self.filter_view.get_filters_from_selection(tree_selection)
        self.selection_filters = filter_strings
        self.task_view.set_filters(self.selection_filters + self.search_filters)

    def filter_text_changed(self, new_text):
        filter_strings = []
        if new_text.strip():
            filter_strings.append(new_text.strip())
        self.search_filters = filter_strings
        self.task_view.set_filters(self.selection_filters + self.search_filters)

    def show_shortcuts_window(self):
        builder = Gtk.Builder()
        d, j = os.path.dirname, os.path.join
        builder.add_from_file(
            simplegtd.libwhiz.path.find_data_file(
                "shortcuts-window.ui",
                "simplegtd",
                before=j(d(d(d(__file__))), "data")
            )
        )
        shortcuts_window = builder.get_object("shortcuts-simplegtd")
        shortcuts_window.set_transient_for(self)
        shortcuts_window.show_all()
        shortcuts_window.set_property("view-name", "")
        shortcuts_window.set_property("section-name", "shortcuts")



#!/usr/bin/python3

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gdk


def get_base_attr(klass, attr):
    '''Retrieves a class attribute from one of the bases of self.

    Arguments:
        klass: The class you want to get the class base attribute from.

    Returns:
        The class attribute in question, or None if not found.
    '''
    for base_ in klass.__bases__:
        if base_ == klass:
            continue
        if not hasattr(base_, attr):
            continue
        return getattr(base_, attr)


class _WindowState(object):

    current_width = None
    current_height = None
    is_maximized = None
    is_fullscreen = None

    @classmethod
    def from_keyfile(klass, filename):
        self = klass()
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(filename, GLib.KeyFileFlags.NONE)
            self.current_width = f.get_integer("window_state", "current_width")
            self.current_height = f.get_integer("window_state", "current_height")
            self.is_maximized = f.get_boolean("window_state", "is_maximized")
            self.is_fullscreen = f.get_boolean("window_state", "is_fullscreen")
        except GLib.Error:
            pass
        return self

    def to_keyfile(self, filename):
        '''Saves the window state to a keyfile.

        The directory containing the filename must already exist.
        '''
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(filename, GLib.KeyFileFlags.NONE)
        except GLib.Error:
            pass
        if self.current_width is not None:
            f.set_integer("window_state", "current_width", self.current_width)
        if self.current_height is not None:
            f.set_integer("window_state", "current_height", self.current_height)
        if self.is_maximized is not None:
            f.set_boolean("window_state", "is_maximized", self.is_maximized)
        if self.is_fullscreen is not None:
            f.set_boolean("window_state", "is_fullscreen", self.is_fullscreen)
        f.save_to_file(filename)


class RememberingWindow(object):
    '''Mixin class for Gtk.Window and Gtk.ApplicationWindow that remembers
    its onscreen position, given a state file to store this info.

    Derived (perhaps imperfectly) from:
    https://wiki.gnome.org/HowDoI/SaveWindowState
    '''

    def __init__(self, state_file):
        '''Initializes the RememberingWindow object.

        Takes a path name where the window state will be saved.
        Per XDG standards, we recommend a file inside the folder
        `simplegtd.libwhiz.path.cache_home(yourappname)`.

        If you call self.set_default_size() after initializing
        this object, this object will not work as advertised. In other
        words: please invoke self.__init__(state_file) after initializing
        the Gtk.Window / Gtk.ApplicationWindow you're mixing this class
        into, and also after any calls to self.set_default_size().
        '''
        if not get_base_attr(self.__class__, "size_allocate"):
            raise TypeError(
                "%s must subclass a Gtk.Window or subclass thereof" % (
                    self
                )
            )

        self.__state_file = state_file
        self.__window_state = _WindowState.from_keyfile(self.__state_file)
        if (self.__window_state.current_width is not None
            and self.__window_state.current_height is not None):
            self.set_default_size(self.__window_state.current_width,
                                  self.__window_state.current_height)
        if self.__window_state.is_maximized:
            self.maximize()
        if self.__window_state.is_fullscreen:
            self.fullscreen()
        self.connect('size-allocate', self.on_size_allocate)
        self.connect('window-state-event', self.on_window_state_event)
        self.connect("destroy", self.on_destroy)

    def on_size_allocate(self, unused_window, allocation):
        size_allocate = get_base_attr(self.__class__, "size_allocate")
        size_allocate(self, allocation)
            
        if (not (self.__window_state.is_maximized or self.__window_state.is_fullscreen)):
            s = self.get_size()
            self.__window_state.current_width = s.width
            self.__window_state.current_height = s.height

    def on_window_state_event(self, unused_window, window_state):
        self.__window_state.is_maximized = (window_state.new_window_state & Gdk.WindowState.MAXIMIZED) != 0
        self.__window_state.is_fullscreen = (window_state.new_window_state & Gdk.WindowState.FULLSCREEN) != 0

    def on_destroy(self, unused_window):
        self.__window_state.to_keyfile(self.__state_file)



import gi
gi.require_version('Pango', '1.0')
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Pango, Gdk, Gtk

import simplegtd.libwhiz.css


class TaskView(Gtk.TreeView):

    filters = None
    filter_displays_empty_tasks = False
    editing_iter = None

    def __init__(self):
        Gtk.TreeView.__init__(self)
        self.set_headers_visible(False)
        self.set_property('expand', True)
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        self.cell = renderer = Gtk.CellRendererText()
        renderer.set_property('editable', True)
        renderer.set_property('wrap-mode', Pango.WrapMode.WORD)
        renderer.set_property('ellipsize', Pango.EllipsizeMode.END)
        renderer.connect("edited", self.task_edited)
        renderer.connect("editing-canceled", self.task_editing_canceled)
        self.column = Gtk.TreeViewColumn("Task", renderer, markup=1)
        self.append_column(self.column)
        self.connect("key-press-event", self.key_press_event)

    def key_press_event(self, unused_widget, event):
        # FIXME if modifier keys are depressed these should probably be ignored.
        if event.keyval in (Gdk.KEY_Delete, Gdk.KEY_KP_Delete):
            filtered_model, selected_paths = self.get_selection().get_selected_rows()
            underlying_model = filtered_model.get_model()
            to_remove = []
            for path in reversed(selected_paths):
                filtered_iter = filtered_model[path].iter
                underlying_iter = filtered_model.convert_iter_to_child_iter(filtered_iter)
                to_remove.append(underlying_iter)
            underlying_model.remove_many(to_remove)
        elif event.keyval in (Gdk.KEY_Insert, Gdk.KEY_KP_Insert, Gdk.KEY_plus):
            self.filter_displays_empty_tasks = True
            self.get_model().refilter()
            # FIXME: implement Shift+Insert to do the same as Ctrl+V
            filtered_model, selected_paths = self.get_selection().get_selected_rows()
            underlying_model = filtered_model.get_model()
            pos = None
            for path in selected_paths:
                filtered_iter = filtered_model[path].iter
                pos = filtered_model.convert_iter_to_child_iter(filtered_iter)
                break
            new_row_iter = underlying_model.new_at(pos)
            self.editing_iter = new_row_iter
            got_set, new_filtered_row_iter = filtered_model.convert_child_iter_to_iter(new_row_iter)
            if got_set:
                self.get_selection().unselect_all()
                self.get_selection().select_iter(new_filtered_row_iter)
                self.scroll_to_cell(filtered_model.get_path(new_filtered_row_iter), self.column)
                self.set_cursor(filtered_model.get_path(new_filtered_row_iter), self.column, True)

    def task_editing_canceled(self, unused_renderer):
        if self.editing_iter:
            if self.get_model().get_model().get_at(self.editing_iter) == "":
                self.get_model().get_model().remove_at(self.editing_iter)
        self.editing_iter = None

    def task_edited(self, unused_renderer, path, new_text):
        filtered_model = self.get_model()
        todotxt_model = filtered_model.get_model()
        filtered_iter = filtered_model[path].iter
        todotxt_iter = filtered_model.convert_iter_to_child_iter(filtered_iter)
        if new_text == "":
            todotxt_model.remove_at(todotxt_iter)
        else:
            todotxt_model.edit(todotxt_iter, new_text)
        self.filter_displays_empty_tasks = False
        self.get_model().refilter()

    def set_model(self, model):
        Gtk.TreeView.set_model(self, model.filter_new())

    def filter_func(self, model, it, unused_ptr):
        task = model.get_value(it, 0).text
        if self.filter_displays_empty_tasks and task == "":
            return True
        pluses = []
        ats = []
        substrings = []
        for flt in self.filters:
            if flt == "":
                continue
            elif flt.startswith("+"):
                pluses.append(flt.lower())
            elif flt.startswith("@"):
                ats.append(flt.lower())
            else:
                substrings.append(flt.lower())
        has_any_pluses = any(i in task.lower() for i in pluses) if pluses else True
        has_any_ats = any(i in task.lower() for i in ats) if ats else True
        has_all_substrings = all(i in task.lower() for i in substrings) if substrings else True
        result = has_any_pluses and has_any_ats and has_all_substrings
        return result

    def set_filters(self, filters):
        if self.filters is None:
            self.filters = filters
            self.get_model().set_visible_func(self.filter_func)
        else:
            self.filters = filters
            self.get_model().refilter()

    def focus_first(self):
        try:
            model = self.get_model()
            self.set_cursor_on_cell(
                model.get_path(model[0].iter),
                self.column,
                self.cell,
                False
            )
        except IndexError: pass


class FilterView(Gtk.TreeView):

    css = """
.filter_view {
    background-color: transparent;
    color: @theme_text_color;
}
.filter_view:selected {
    color: @theme_selected_fg_color;
    background-color: @theme_selected_bg_color;
}
"""

    def __init__(self):
        Gtk.TreeView.__init__(self)
        self.set_headers_visible(False)
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        renderer = Gtk.CellRendererText()
        renderer.ellipsize = Pango.EllipsizeMode.MIDDLE
        c = Gtk.TreeViewColumn("Filters", renderer, markup=3)
        c.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.append_column(c)
        simplegtd.libwhiz.css.add(self.css)
        self.get_style_context().add_class("filter_view")

    def set_model(self, model):
        Gtk.TreeView.set_model(self, model.get_sorted())

    def get_filters_from_selection(self, tree_selection):
        filter_strings = []
        sorted_model, paths = tree_selection.get_selected_rows()
        underlying_model = sorted_model.get_model()
        for path in paths:
            sorted_iter = sorted_model[path].iter
            underlying_iter = sorted_model.convert_iter_to_child_iter(sorted_iter)
            filter_strings.append(underlying_model.get_filter_string(underlying_iter))
        return filter_strings

    def select_first(self):
        try:
            model = self.get_model()
            self.get_selection().select_iter(model[0].iter)
        except IndexError: pass



import re

from gi.repository import GLib


# FIXME in lieu of using a CellRenderer for a task object,
# we will temporarily use this shit.
def markup_for(task_text):
    date_regex = re.compile("^([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))$")
    list_regex = re.compile("^@[a-zA-Z0-9]")
    context_regex = re.compile("^@@[a-zA-Z0-9]")
    prio_regex = re.compile("^[(][A-Z][)]$")
    tag_regex = re.compile("^[+]")
    newtoks = []
    m = GLib.markup_escape_text
    for tok in task_text.split(" "):
        if date_regex.match(tok):
            # FIXME if date is overdue mark it up as red.
            newtoks.append("<span foreground='green'>%s</span>" % m(tok))
        elif context_regex.match(tok):
            newtoks.append("<span foreground='darkblue'>%s</span>" % m(tok))
        elif list_regex.match(tok):
            newtoks.append("<span foreground='darkgreen'>%s</span>" % m(tok))
        elif tag_regex.match(tok):
            newtoks.append("<span foreground='darkviolet'>%s</span>" % m(tok))
        elif prio_regex.match(tok):
            newtoks.append("<span foreground='chocolate'>%s</span>" % m(tok))
        else:
            newtoks.append(m(tok))
    new_text = " ".join(newtoks)
    if task_text.startswith("x "):
        new_text = "<span strikethrough='true' color='gray'>%s</span>" % new_text
    return new_text



import errno
import logging
import os
import threading

import gi
gi.require_version('GObject', '2.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gio


class AbstractBlobStore(GObject.GObject):
    '''An AbstractBlobStore is an object type that presents an interface to
    views, which are interested in data from a source (e.g. a file, or a block
    of memory, and want to obtain its contents, as well as get notified in an
    asynchronous manner when the data is being obtained.

    The BlobStore emits several signals:

    1. `reading`, when the source has changed, and the BlobStore is about
       to re-read the source in order to obtain the data.
    2. `done-reading`, when the BlobStore is finished reading the source.
       Two parameters accompany the signal:
       a) the data if successfully read (None if unsuccessful).
       b) the exception that happened, if unsuccessful (none, if successful).
    3. `writing`, when the object has been told to save the data to the
       source via put().
    4. `done-writing`, when the object is done saving the data to the source.
       One parameter accompanies the signal:
       a) the data that was just written.
       a) None if the save was successful, an exception if the save failed.

    Subscribers should subscribe to both signals to know when it is safe to
    call store() -- subscribers should not call store() in between the emission
    of the `reading` and the `done-reading` signal.
    '''

    __gsignals__ = {
        'reading': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'done-reading': (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
        'writing': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'done-writing': (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

    def open(self, source):
        '''Opens the source, creating if it does not exist.  Immediately after
        calling open(), the signals `reading` and then `done-reading` will be
        emitted.

        The meaning of the `source` parameter depends upon the implementation
        of the subclass.

        Emits an exception (via the `done-reading` signal) if it could not
        open or create the source.

        If an exception is emitted, the caller should assume this object is
        invalid, disconnect all the connections it made, and discard this
        object altogether.

        You cannot open() an object twice.
        '''
        raise NotImplementedError

    def close(self):
        '''Closes the source, finishing monitoring of file changes.

        No signals will be raised after closing the source.  It is an error
        to call put() after close().

        Raises an exception if it could not open the source and it could also
        not create the source (if it didn't exist).
        '''
        raise NotImplementedError

    def put(self, data):
        '''Store data in the source.

        Emits an exception (via the `done-writing` signal) if it could not
        store the data successfully.  If an exception is emitted, unless
        it is an IOError(EBUSY), the caller should assume further tries to
        put() data to the object will fail.

        It is a ValueError to call put() in between the `reading` and
        `done-reading` signals.
        '''
        raise NotImplementedError


class StringBlobStore(AbstractBlobStore):

    memory = None
    __closed = False

    def open(self, source):
        (
            AbstractBlobStore.open.__doc__,
            '''

        This implementation uses the string representation of the `source`
        parameter to initialize its memory.
        '''
        )
        if self.memory is not None:
            raise ValueError("open() operation on opened %s" % self.__class__.__name__)
        if self.__closed:
            raise ValueError("open() operation on closed %s" % self.__class__.__name__)
        self.memory = str(source) or ''
        self.emit('reading')
        self.emit('done-reading', self.memory, None)

    def close(self):
        self.__closed = True
        self.memory = None

    def put(self, data):
        (
            AbstractBlobStore.put.__doc__,
            '''

        This implementation stores the string representation of the `data`
        parameter in its memory.
        '''
        )
        if self.memory is None:
            raise ValueError("put() operation on unopened %s" % self.__class__.__name__)
        if self.__closed:
            raise ValueError("put() operation on closed %s" % self.__class__.__name__)
        self.emit("writing")
        self.memory  = str(data)
        self.emit("done-writing", data, None)


class FileBlobStore(AbstractBlobStore):

    name = None
    __closed = False
    _last_changed = (-1, -1, -1)
    __file_changes_detected = 0
    __monitor_dir = None
    __monitor_handle = None
    __loading = False

    def __init__(self):
        AbstractBlobStore.__init__(self)
        self.logger = logging.getLogger(self.__class__.__name__)

    def open(self, source):
        (
            AbstractBlobStore.open.__doc__,
            '''

        This implementation uses a file named in the `source` parameter,
        creating the file if it does not exist.
        '''
        )
        if self.name is not None:
            raise ValueError("open() operation on opened %s" % self.__class__.__name__)
        if self.__closed:
            raise ValueError("open() operation on closed %s" % self.__class__.__name__)

        self.name = source
        self.__begin_watching()
        self._load()

    def __begin_watching(self):
        assert not self.__monitor_handle
        giofile_dir = Gio.File.new_for_path(os.path.dirname(self.name))
        self.__monitor_dir = giofile_dir.monitor(Gio.FileMonitorFlags.WATCH_MOVES, None)
        self.__monitor_handle = self.__monitor_dir.connect("changed", self.__dir_changed)

    def __pause_watching(self):
        self.__monitor_dir.disconnect(self.__monitor_handle)
        self.__monitor_handle = None
        self.__monitor_dir.cancel()
        self.__monitor_dir = None

    def __dir_changed(self, unused_monitor, unused_f, newf, event):
        if self.__closed:
            return
        if self.__loading:
            return
        self.__file_changes_detected += 1
        if event == Gio.FileMonitorEvent.CHANGES_DONE_HINT:
            self.logger.debug("Directory changed.  Reloading.")
            self._load()
        elif event == Gio.FileMonitorEvent.RENAMED:
            if newf.get_basename() == os.path.basename(self.name):
                self.logger.debug("A file was renamed to %s.  Reloading." % self.name)
                self._load()

    # FIXME: block edits while loading.
    # And block edits while saving.
    def _load(self, previously_read_data=False):
        if self.__closed:
            # File was closed in between speculative loads.
            # Simply ignore this.
            return False
        if previously_read_data is False:
            self.__loading = True
            self.emit("reading")
            self.logger.debug("About to read %s" % self.name)

        try:
            try:
                with open(self.name, "r") as f:
                    statf = os.stat(self.name)
                    data = f.read()
            except FileNotFoundError:
                with open(self.name, "w"):
                    pass

            if previously_read_data is False:
                # We're going to wait a bit and reload, mainly to
                # see if the file's contents are stable over some time.
                self.logger.debug('Waiting a bit to re-read %s before we declare success.', self.name)
                GLib.timeout_add(75, lambda: self._load(data) and False)
            elif data == previously_read_data:
                # The file's contents have remained stable.
                # We can truly load now.
                self.logger.debug('File %r is stable now, truly reloading.', self.name)
                self._last_changed = (getattr(statf, "st_mtime_ns", statf.st_mtime), getattr(statf, "st_ctime_ns", statf.st_ctime), statf.st_size)
                self.__loading = False
                self.logger.debug("Done reading %s" % self.name)
                self.emit("done-reading", data, None)
                return False
            else:
                # Uh-oh, the file changed from under us.
                # We must once again re-run this.
                self.logger.debug('File %r changed from under us while we were loading, waiting again.', self.name)
                GLib.timeout_add(75, lambda: self._load(data) and False)

        except Exception as e:
            self.__loading = False
            self.logger.debug("Error reading %s: %s" % (self.name, e))
            self.emit("done-reading", None, e)

    def close(self):
        if self.__monitor_handle:
            self.__monitor_dir.disconnect(self.__monitor_handle)
        if self.__monitor_dir:
            self.__monitor_dir.cancel()
        self.__closed = True
        self.name = None

    def put(self, data):
        (
            AbstractBlobStore.put.__doc__,
            '''

        This implementation stores the string representation of the `data`
        parameter in the file named in open().

        If the file is in the middle of reading because it changed, `done-reading`
        emits IOError(errno.EAGAIN).  Caller must attempt to put() again later,
        or preferably cancel/undo the operation that started the put(), because
        data loss is likely (overwrite of out-of-date data).
        '''
        )
        if self.name is None:
            raise ValueError("put() operation on unopened %s" % self.__class__.__name__)
        if self.__closed:
            raise ValueError("put() operation on closed %s" % self.__class__.__name__)
        if self.__closed:
            raise ValueError("put() operation on closed %s" % self.__class__.__name__)
        if self.__loading:
            raise IOError(errno.EAGAIN, "put() operation on busy %s" % self.__class__.__name__)

        self.__pause_watching()
        self.emit("writing")
        self.logger.debug("About to write %s" % self.name)

        def _finish_put(name, new_data):
            try:
                with open(name, "w+") as f:
                    if new_data == f.read():
                        self.logger.debug("Unchanged.  No writes.")
                    else:
                        f.seek(0)
                        f.truncate()
                        f.write(new_data)
                        f.flush()
                self.logger.debug("Done writing %s" % name)
                l1 = lambda: self.emit("done-writing", new_data, None)
                l2 = lambda: self.__begin_watching()
                GObject.idle_add(l1)
                GObject.idle_add(l2)
            except Exception as e:
                self.logger.debug("Error writing %s: %s" % (self.name, e))
                l1 = lambda: self.emit("done-writing", None, e)
                l2 = lambda: self.__begin_watching()
                GObject.idle_add(l1)
                GObject.idle_add(l2)

        t = threading.Thread(target=_finish_put, args=(self.name, data))
        t.start()



#!/usr/bin/python3

import difflib
import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk


# FIXME it's probably prudent to create a Mapper between the LineStore
# and the filesystem, so that I can move the hairy load/save code
# outside of this, and then the interface between the file system
# and the LineStore is just lines, no more than that.
# Perhaps it becomes even possible to consolidate the TaskStore and
# the ListStore completely.  Fundamentally the concern of handling the
# file system stuff and the one of handling the actual in-memory repr
# of the line/task store should not be mixed in one single class.
class LineStore(Gtk.ListStore):

    last_line_cr = False
    __watches = None

    __gsignals__ = {
        'unserialized': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'serialized': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self):
        Gtk.ListStore.__init__(self, str)
        self.logger = logging.getLogger(self.__class__.__name__)

    def unserialize(self, unused_bs, text, exc):
        self.logger.debug("Load begun.  Lines: %s.", len(text.splitlines()))
        if exc is not None:
            # Oops.  The blobstore had a problem.
            # Our caller deals with these situations.  We do not know how.
            return

        new_lines = text.splitlines(True)
        existing_lines = [t[0] for t in self]
        self.last_line_cr = new_lines and new_lines[-1] and new_lines[-1][-1] == "\n"
        def strip_cr(line):
            if line and line[-1] == "\n":
                line = line[:-1]
            return line
        new_lines = [strip_cr(l) for l in new_lines]
        if new_lines and not new_lines[-1]:
            new_lines = new_lines[:-1]

        diff = difflib.SequenceMatcher()
        diff.set_seqs(existing_lines, new_lines)
        added = 0
        removed = 0
        changed = 0
        for op, i1, i2, j1, j2 in diff.get_opcodes():
            if op == 'equal':
                pass
            elif op == 'insert':
                r = 0
                for row in new_lines[j1:j2]:
                    #self.logger.debug("Inserting row %s", i1 + r)
                    self.insert(i1 + r, row=[row])
                    r = r + 1
                    added += 1
            elif op == 'delete':
                for r in reversed(range(i1, i2)):
                    #self.logger.debug("Removing row %s", r)
                    self.remove(self[r].iter)
                    removed += 1
            elif op == 'replace':
                if i2 - i1 == j2 - j1:
                    # The number of rows replaced is the same in old and new.
                    for r in range(i2 - i1):
                        #self.logger.debug("Replacing row %s with new row %s", i1 + 1, j1 + r)
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        changed += 1
                elif i2 - i1 > j2 - j1:
                    # The number of rows replaced is larger in old than in new.
                    for r in reversed(range((i2 - i1) - (j2 - j1))):
                        #self.logger.debug("Removing row %s", i1 + r)
                        self.remove(self[i1+r].iter)
                        removed += 1
                    for r in range(j2 - j1):
                        #self.logger.debug("Replacing row %s with new row %s", i1 + 1, j1 + r)
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        changed += 1
                else: # i2 - i1 < j2 - j1:
                    # The number of rows replaced is larger in new than in old.
                    for r in range((j2 - j1) - (i2 - i1)):
                        #self.logger.debug("Inserting row %s  %r", i1 + r, new_lines[j1 + r])
                        self.insert(i1 + r, [new_lines[j1 + r]])
                        added += 1
                    for r in range(j2 - j1):
                        #self.logger.debug("Replacing row %s with new row %s  %r", i1 + 1, j1 + r, new_lines[j1 + r])
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        changed += 1
            else:
                assert 0, "not reached: %s" % op
        self.logger.debug("Load finished.  %s additions, %s removals, %s changes.  Lines: %s.", added, removed, changed, len(self))
        self.emit('unserialized')

    def serialize(self):
        lines = [row[0] + "\n" for row in self]
        if not self.last_line_cr:
            if lines:
                lines[-1] = lines[-1][:-1]
        text = "".join(lines)
        self.emit('serialized')
        return text



'''See https://wiki.gnome.org/HowDoI/CustomStyle'''

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, Gtk


__css_adjusted = {}


def add(css):
    # Must only ever be called at runtime, not at import time.
    # See https://github.com/GNOME/gtk/blob/master/gtk/theme/Adwaita/_colors-public.scss
    global __css_adjusted
    if css not in __css_adjusted:
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(css.encode("utf-8"))
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
    __css_adjusted[css] = True



#!/usr/bin/python3

import glob
from setuptools import setup
import os
import platform

dir = os.path.dirname(__file__)
path_to_main_file = os.path.join(dir, "src/simplegtd/__init__.py")
path_to_readme = os.path.join(dir, "README.md")
for line in open(path_to_main_file):
	if line.startswith('__version__'):
		version = line.split()[-1].strip("'").strip('"')
		break
else:
	raise ValueError('"__version__" not found in "src/simplegtd/__init__.py"')
readme = open(path_to_readme).read(-1)

classifiers = [
'Development Status :: 3 - Alpha',
'Environment :: X11 Applications :: GTK',
'Intended Audience :: End Users/Desktop',
'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
'Operating System :: POSIX :: Linux',
'Programming Language :: Python :: 3 :: Only',
'Programming Language :: Python :: 3.6',
'Topic :: Office/Business :: Scheduling',
'Topic :: Office/Business :: Groupware',
]

programs = [
    "simplegtd",
]

ui_files = [
    "shortcuts-window.ui",
]

icons = [
    "simplegtd.svg",
]

# Don't write to /usr/share/applications on OS X to work around the
# 'System Integrity Protection'.
g = lambda p: os.path.join(os.path.dirname(__file__), p)
data_files = [
	("/usr/share/applications", [g("applications/%s.desktop") % p for p in programs]),
	("/usr/share/icons", [g("icons/%s") % p for p in icons]),
	("/usr/share/simplegtd", [g("data/%s") % p for p in ui_files]),
] if platform.system() != 'Darwin' else []

setup(
	name='simplegtd',
	version=version,
	description='Manage your to-do list using the Getting Things Done system.',
	long_description = readme,
	author='Manuel Amador (Rudd-O)',
	author_email='rudd-o@rudd-o.com',
	license="GPLv3+",
	url='https://pypi.org/project/simplegtd/',
	package_dir=dict([
                    ("simplegtd", "src/simplegtd"),
					]),
	classifiers = classifiers,
	packages=["simplegtd", "simplegtd.libwhiz", "simplegtd.models"],
	data_files = data_files,
	scripts=["bin/%s" % p for p in programs],
	keywords="gtd getting things done TODO.TXT",
	zip_safe=False,
	install_requires=['pyxdg'],
)



import hashlib
import os

import xdg.BaseDirectory


def find_data_file(filename, app, before=None):
    '''Searches for filename in os.path.join(datadir, app) for datadir
    in XDG_DATA_DIRS.  If before is not None, it looks for the file in that
    directory first.

    Raises:
        KeyError: if filename is not found.
    '''
    for path in (
        ([before] if before is not None else [])
        + [os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, "data")]
        + [os.path.join(x, "simplegtd") for x in xdg.BaseDirectory.xdg_data_dirs]
    ):
        f = os.path.join(path, os.path.basename(filename))
        if os.path.isfile(f):
            return f
    raise KeyError(f)


def config_home(subdir=None):
    if subdir is None:
        return xdg.BaseDirectory.xdg_config_home
    return os.path.join(xdg.BaseDirectory.xdg_config_home, subdir)


def data_home(subdir=None):
    if subdir is None:
        return xdg.BaseDirectory.xdg_data_home
    return os.path.join(xdg.BaseDirectory.xdg_data_home, subdir)


def cache_home(subdir=None):
    if subdir is None:
        return xdg.BaseDirectory.xdg_cache_home
    return os.path.join(xdg.BaseDirectory.xdg_cache_home, subdir)


def strip_data_home(path):
    '''Given a path (relative or absolute), strips the beginning of the path
    if it begins with `data_home()`, else returns the same path unchanged.'''
    abspath = os.path.abspath(path)
    if abspath.startswith(data_home() + "/") or abspath == data_home():
        return abspath[len(data_home()) + 1:]
    return path


def abbrev_home(filename):
    '''Given a path (relative or absolute), changes the path to use ~/
    notation if it is within $HOME, else returns the same path unchanged.'''
    absfilename = os.path.abspath(filename)
    if absfilename.startswith(os.path.expanduser("~/")):
        return "~/" + absfilename[len(os.path.expanduser("~/")):]
    return filename


def hash_path(filename):
    '''Returns an MD5 sum of a file name path.  Or any string, really.'''
    h = hashlib.md5()
    h.update(filename.encode("utf-8", "ignore"))
    h = h.hexdigest()
    return h



import contextlib
import os
import tempfile
import threading
import time
import unittest

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib

import simplegtd.models.blobstore as bs
import simplegtd.models.linestore as sm


def fixture_path(name):
    return os.path.join(os.path.dirname(__file__), "fixtures", name)


def fixture_text(name):
    with open(fixture_path(name), "r") as f:
        return f.read()


def fixture_list(name):
    return [x for x in fixture_text(name).splitlines()]


def dump(liststore):
    elements = []
    for row in liststore:
        elements.append(row[0])
    return elements


@contextlib.contextmanager
def ctx(kallable):
    obj = kallable()
    try:
        yield obj
    finally:
        if hasattr(obj, "close"):
            obj.close()


def transform_file_text(fobject, before_list, after_list, flush):
    assert len(before_list) <= len(after_list)
    working_list = list(before_list)
    fobject.seek(0)
    fobject.truncate(0)
    fobject.write("\n".join(working_list))
    yield 0
    for n, (new) in enumerate(after_list):
        try:
            working_list[n] = new
        except IndexError:
            working_list.append(new)
        fobject.seek(0)
        fobject.truncate(0)
        fobject.write("\n".join(working_list))
        if flush:
            fobject.flush()
        yield n + 1
    assert working_list == after_list


def gloop():
    store = []
    c = GLib.MainContext.default()
    collector = lambda *a: store.append(a)
    def evac(pos):
        while c.pending() or not store:
            c.iteration(False)
        val = store[0][pos]
        store.clear()
        return val
    return collector, evac


def check_exc(datum):
        if isinstance(datum, Exception):
            raise datum


class LineStoreTest(unittest.TestCase):

    def test_loads_fine_from_string_blobstore(self):
        with ctx(lambda: bs.StringBlobStore()) as b:
            with ctx(lambda: sm.LineStore()) as t:
                b.connect("done-reading", t.unserialize),
                b.open(fixture_text("before.txt"))
                gloop()
                loaded = dump(t)
                read = fixture_list("before.txt")
                self.assertListEqual(loaded, read)

    def test_loads_fine(self):
        with ctx(lambda: bs.FileBlobStore()) as b:
            with ctx(lambda: sm.LineStore()) as t:
                b.connect("done-reading", t.unserialize),
                collect, result = gloop()
                b.connect("done-reading", collect)
                b.open(fixture_path("before.txt"))
                check_exc(result(2))
                self.assertListEqual(dump(t), fixture_list("before.txt"))

    def test_loads_written_text_from_another_process(self):
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            with ctx(lambda: bs.FileBlobStore()) as b:
                with ctx(lambda: sm.LineStore()) as t:
                    b.connect("done-reading", t.unserialize),
                    collect, result = gloop()
                    self.assertListEqual(dump(t), [])
                    b.connect("done-reading", collect)
                    b.open(fixture_path("before.txt"))
                    f.write(fixture_text("before.txt"))
                    f.flush()
                    check_exc(result(2))
                    self.assertListEqual(dump(t), fixture_list("before.txt"))

    def _test_transform_line_by_line_in_mainloop(self, flush):
        before_list = fixture_list("before.txt")
        after_list = fixture_list("after.txt")
        def file_writer(f, flush):
            def target():
                for unused_linenumber in transform_file_text(f, before_list, after_list, flush):
                    time.sleep(0.001)
            return threading.Thread(target=target)

        with tempfile.TemporaryDirectory() as tmpdir:
            with tempfile.NamedTemporaryFile(mode="w+", dir=tmpdir) as f:
                with ctx(lambda: bs.FileBlobStore()) as b:
                    with ctx(lambda: sm.LineStore()) as t:
                        b.connect("done-reading", t.unserialize),
                        th = file_writer(f, flush)
                        c = GLib.MainContext.default()
                        b.open(f.name)
                        th.start()
                        while c.pending() or dump(t) != after_list:
                            c.iteration(False)
                            time.sleep(0.05)
                        th.join()
                        x = os.stat(f.name)
                        last_changed = (getattr(x, "st_mtime_ns", x.st_mtime), getattr(x, "st_ctime_ns", x.st_ctime), x.st_size)
                        self.assertListEqual(
                            dump(t), after_list,
                            "Comparison failed.  Object last changed: %s.  Actual last changed: %s." % (
                                b._last_changed,
                                last_changed
                            )
                        )

    def test_transform_line_by_line_in_mainloop_with_flushing(self):
        return self._test_transform_line_by_line_in_mainloop(True)

    def test_transform_line_by_line_in_mainloop_without_flushing(self):
        return self._test_transform_line_by_line_in_mainloop(False)
